import os

import click

from ..settings import LOCAL_CONFIG, load_config
from ..utils import taskstatus
from ..helpers import get_issue, get_transition, set_issue_status, assign_issue
from ..helpers import get_branch, get_project_keys, get_project_dirs


@click.command()
@click.option('-k', '--key', 'method', flag_value='key', default=True)
@click.option('-d', '--dir', 'method', flag_value='dir')
@click.argument('project')
@click.pass_obj
def activate(lancet, method, project):
    """Switch to this project."""
    with taskstatus('Looking up project') as ts:
        if method == 'key':
            func = get_project_keys
        elif method == 'dir':
            func = get_project_keys

        for key, project_path in func(lancet):
            if key.lower() == project.lower():
                break
        else:
            ts.abort('Project "{}" not found (using {}-based lookup)',
                     project, method)

    # Load the configuration
    config = load_config(os.path.join(project_path, LOCAL_CONFIG))

    # cd to the project directory
    lancet.defer_to_shell('cd', project_path)

    # Activate virtualenv
    venv = config.get('lancet', 'virtualenv', fallback=None)
    if venv:
        venv_path = os.path.join(project_path, os.path.expanduser(venv))
        activate_script = os.path.join(venv_path, 'bin', 'activate')
        lancet.defer_to_shell('source', activate_script)
    else:
        if 'VIRTUAL_ENV' in os.environ:
            lancet.defer_to_shell('deactivate')


@click.command()
@click.pass_obj
def _project_keys(lancet):
    click.echo('\n'.join(k.lower() for k, p in get_project_keys(lancet)))


@click.command()
@click.pass_obj
def _project_dirs(lancet):
    click.echo('\n'.join(d.lower() for d, p in get_project_dirs(lancet)))


@click.command()
@click.option('--base', '-b', 'base_branch',
              help='Base branch to branch off from.')
@click.argument('issue')
@click.pass_context
def workon(ctx, issue, base_branch):
    """
    Start work on a given issue.

    This command retrieves the issue from the issue tracker, creates and checks
    out a new aptly-named branch, puts the issue in the configured active,
    status, assigns it to you and starts a correctly linked Harvest timer.

    If a branch with the same name as the one to be created already exists, it
    is checked out instead. Variations in the branch name occuring after the
    issue ID are accounted for and the branch renamed to match the new issue
    summary.

    If the `default_project` directive is correctly configured, it is enough to
    give the issue ID (instead of the full project prefix + issue ID).
    """
    lancet = ctx.obj

    username = lancet.config.get('tracker', 'username')
    active_status = lancet.config.get('tracker', 'active_status')
    if not base_branch:
        base_branch = lancet.config.get('repository', 'base_branch')

    # Get the issue
    issue = get_issue(lancet, issue)

    # Get the working branch
    branch = get_branch(lancet, issue, base_branch)

    # Make sure the issue is in a correct status
    transition = get_transition(ctx, lancet, issue, active_status)

    # Make sure the issue is assigned to us
    assign_issue(lancet, issue, username, active_status)

    # Activate environment
    set_issue_status(lancet, issue, active_status, transition)

    with taskstatus('Checking out working branch') as ts:
        lancet.repo.checkout(branch.name)
        ts.ok('Checked out working branch based on "{}"'.format(base_branch))

    with taskstatus('Starting harvest timer') as ts:
        lancet.timer.start(issue)
        ts.ok('Started harvest timer')


@click.command()
@click.argument('issue', required=False)
@click.pass_obj
def time(lancet, issue):
    """
    Start an Harvest timer for the given issue.

    This command takes care of linking the timer with the issue tracker page
    for the given issue. If the issue is not passed to command it's taken
    from currently active branch.
    """
    issue = get_issue(lancet, issue)

    with taskstatus('Starting harvest timer') as ts:
        lancet.timer.start(issue)
        ts.ok('Started harvest timer')


@click.command()
@click.pass_context
def pause(ctx):
    """
    Pause work on the current issue.

    This command puts the issue in the configured paused status and stops the
    current Harvest timer.
    """
    lancet = ctx.obj
    paused_status = lancet.config.get('tracker', 'paused_status')

    # Get the issue
    issue = get_issue(lancet)

    # Make sure the issue is in a correct status
    transition = get_transition(ctx, lancet, issue, paused_status)

    # Activate environment
    set_issue_status(lancet, issue, paused_status, transition)

    with taskstatus('Pausing harvest timer') as ts:
        lancet.timer.pause()
        ts.ok('Harvest timer paused')


@click.command()
@click.pass_context
def resume(ctx):
    """
    Resume work on the currently active issue.

    The issue is retrieved from the currently active branch name.
    """
    lancet = ctx.obj

    username = lancet.config.get('tracker', 'username')
    active_status = lancet.config.get('tracker', 'active_status')

    # Get the issue
    issue = get_issue(lancet)

    # Make sure the issue is in a correct status
    transition = get_transition(ctx, lancet, issue, active_status)

    # Make sure the issue is assigned to us
    assign_issue(lancet, issue, username, active_status)

    # Activate environment
    set_issue_status(lancet, issue, active_status, transition)

    with taskstatus('Resuming harvest timer') as ts:
        lancet.timer.start(issue)
        ts.ok('Resumed harvest timer')



import os
import glob

import click

from .settings import LOCAL_CONFIG, load_config
from .git import BranchGetter
from .utils import taskstatus


def get_issue(lancet, key=None):
    with taskstatus('Looking up issue on the issue tracker') as ts:
        issue = lancet.get_issue(key)
        summary = issue.fields.summary
        crop = len(summary) > 40
        if crop:
            summary = summary[:40] + '...'
        ts.ok('Retrieved issue {}: {}'.format(issue.key, summary))
    return issue


def get_transition(ctx, lancet, issue, to_status):
    current_status = issue.fields.status.name
    if current_status != to_status:
        transitions = [t['id'] for t in lancet.tracker.transitions(issue)
                       if t['to']['name'] == to_status]
        if not transitions:
            click.secho(
                'No transition from "{}" to "{}" found, aborting.'
                .format(current_status, to_status),
                fg='red', bold=True
            )
            ctx.exit(1)
        elif len(transitions) > 1:
            click.secho(
                'Multiple transitions found from "{}" to "{}", aborting.'
                .format(current_status, to_status),
                fg='red', bold=True
            )
            ctx.exit(1)
        else:
            transition_id = transitions[0]
    else:
        transition_id = None
    return transition_id


def set_issue_status(lancet, issue, to_status, transition):
    with taskstatus('Setting issue status to "{}"'.format(to_status)) as ts:
        if transition is not None:
            lancet.tracker.transition_issue(issue, transition)
            ts.ok('Issue status set to "{}"'.format(to_status))
        else:
            ts.ok('Issue already "{}"'.format(to_status))


def assign_issue(lancet, issue, username, active_status=None):
    with taskstatus('Assigning issue to you') as ts:
        assignee = issue.fields.assignee
        if not assignee or assignee.name != username:
            if issue.fields.status.name == active_status:
                ts.abort('Issue already active and not assigned to you')
            else:
                lancet.tracker.assign_issue(issue, username)
                ts.ok('Issue assigned to you')
        else:
            ts.ok('Issue already assigned to you')


def get_branch(lancet, issue, base_branch=None, create=True):
    if not base_branch:
        base_branch = lancet.config.get('repository', 'base_branch')
    remote_name = lancet.config.get('repository', 'remote_name')

    remote = lancet.repo.lookup_remote(remote_name)
    credentials = lancet.repo.get_credentials_for_remote(remote)

    name_getter = lancet.get_instance_from_config(
        'repository', 'branch_name_getter')
    branch_getter = BranchGetter(base_branch, credentials, name_getter,
                                 remote_name)

    return branch_getter(lancet.repo, issue, create=create)


def get_project_keys(lancet):
    workspace = os.path.expanduser(lancet.config.get('lancet', 'workspace'))
    config_files = glob.glob(os.path.join(workspace, '*', LOCAL_CONFIG))

    for path in config_files:
        config = load_config(path)
        key = config.get('tracker', 'default_project', fallback=None)
        if key:
            yield key, os.path.dirname(path)


def get_project_dirs(lancet):
    workspace = os.path.expanduser(lancet.config.get('lancet', 'workspace'))
    for path in glob.glob(os.path.join(workspace, '*', '.lancet')):
        path = os.path.dirname(path)
        yield os.path.basename(path), path



import click

from lancet.helpers import get_issue
from lancet.utils import taskstatus, edit_template


VERSION_FILTER_QUERY = (
    'project = {project_key} and fixVersion = {version_name}')


def get_version(lancet, project_key, version_name):
    project = lancet.tracker.project(project_key)
    versions = lancet.tracker.project_versions(project)
    for v in versions:
        if v.name == version_name:
            return v


def get_issues_fixed_in_version(lancet, project_key, version_name):
    return lancet.tracker.search_issues(VERSION_FILTER_QUERY.format(
        project_key=project_key, version_name=version_name))


@click.command()
@click.option('-t', '--target',
              help='Target of the release (either a commit or a branch name).')
@click.option('-d', '--draft/--no-draft', default=True,
              help='Creates the release in draft mode (default to true).')
@click.option('-p', '--prerelease/--no-prerelease', default=False,
              help='Creates a prerelease.')
@click.option('-o', '--open/--no-open', 'open_link', default=False,
              help='Opens the link with the release.')
@click.argument('version_name', metavar='version')
@click.pass_obj
def release_notes(lancet, version_name, target, draft, prerelease, open_link):
    project_key = lancet.config.get('tracker', 'default_project')

    with taskstatus('Getting version') as ts:
        version = get_version(lancet, project_key, version_name)
        if not version:
            ts.abort('Version {} not found for project {}',
                     version_name, project_key)

        ts.ok('Got version {}', version.name)

    with taskstatus('Getting issues fixed in version {}', version.name) as ts:
        issues = get_issues_fixed_in_version(lancet, project_key, version_name)
        ts.ok('Found {} issues', len(issues))

    with taskstatus('Creating release') as ts:
        notes = edit_template(
            lancet.config.get('tracker', 'release_notes_template'),
            issues=issues,
            version=version
        )
        if not notes:
            ts.abort('No release notes provided')

        name, notes = notes.split('\n\n', 1)

        if target is None:
            target = lancet.config.get('repository', 'base_branch')

        release = lancet.github_repo.create_release(
            version_name,
            target_commitish=target,
            name=name.strip(),
            body=notes.strip(),
            draft=draft,
            prerelease=prerelease,
        )

        ts.ok('Release created at {}', release.html_url)

    if open_link:
        click.launch(release.html_url)


# TODO: Add commands to:
# * Release versions (both GH + Jira)
# * Set fixVersion to closed issues in the current sprint
# * ...
#
# @click.command()
# @click.argument('version')
# @click.argument('issues', nargs=-1, required=True)
# @click.pass_obj
# def set_version(lancet, version, issues):
#     for issue_key in issues:
#         set_issue_version(lancet, issue_key, version)
#
#
# def set_issue_version(lancet, issue_key, version):
#     issue = get_issue(lancet, issue_key)
#     issue.update(fields={
#         'fixVersion': [{'add': [{'id': 10301}]}]
#     })



import os
import sys
import pdb
import importlib
import shlex
import subprocess

import click
from click.utils import make_str

from . import __version__
from .settings import load_config, PROJECT_CONFIG
from .base import Lancet, WarnIntegrationHelper, ShellIntegrationHelper
from .utils import hr


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class SubprocessExecuter(click.BaseCommand):
    def parse_args(self, ctx, args):
        ctx.args = args
        return args

    def invoke(self, ctx):
        ctx.exit(subprocess.call(ctx.args[0], shell=True))


class ConfigurableLoader(click.Group):

    _config = None
    _path_setup_complete = False

    def __init__(self, *args, **kwargs):
        self.__class__.setup_path()
        super().__init__(*args, **kwargs)

    @classmethod
    def setup_path(cls):
        if not cls._path_setup_complete:
            paths = cls.get_config().getlist('lancet', 'add_to_path')
            for p in reversed(paths):
                sys.path.insert(0, os.path.expanduser(p))

    @classmethod
    def get_config(cls):
        if not cls._config:
            if os.path.exists(PROJECT_CONFIG):
                cls._config = load_config(PROJECT_CONFIG)
            else:
                cls._config = load_config()
        return cls._config

    @classmethod
    def get_configured_commands(cls, config=None):
        if config is None:
            config = cls.get_config()
        return config.options('commands')

    @classmethod
    def get_configured_aliases(cls, config=None):
        if config is None:
            config = cls.get_config()
        return config.options('alias')

    @staticmethod
    def show_help_all(ctx, param, value):
        if value and not ctx.resilient_parsing:
            # Explicitly set the flag to show hidden subcommands, otherwise
            # the code in the `get_help` method would set it to False.
            ctx.show_hidden_subcommands = True
            click.echo(ctx.get_help())
            ctx.exit()

    def list_commands(self, ctx):
        commands = set(super().list_commands(ctx))
        commands = commands.union(self.get_configured_commands())

        # Do not list hidden subcommands if the flag is explicitly set on the
        # context. By default include all commands.
        if not getattr(ctx, 'show_hidden_subcommands', True):
            commands = [c for c in commands if not c.startswith('_')]
        return sorted(commands)

    def list_aliases(self, ctx):
        return sorted(self.get_configured_aliases())

    def get_help(self, ctx):
        # By default do not list hidden subcommands.
        ctx.show_hidden_subcommands = getattr(
            ctx, 'show_hidden_subcommands', False)
        return super().get_help(ctx)

    def format_options(self, ctx, formatter):
        super().format_options(ctx, formatter)
        self.format_aliases(ctx, formatter)

    def format_aliases(self, ctx, formatter):
        rows = []
        for alias in self.list_aliases(ctx):
            rows.append((alias, self.get_config().get('alias', alias)))

        if rows:
            with formatter.section('Aliases'):
                formatter.write_dl(rows)

    def resolve_command(self, ctx, args):
        cmd_name = make_str(args[0])

        if cmd_name in self.get_configured_aliases():
            if cmd_name in self.list_commands(ctx):
                # Shadowing of existing commands is explicitly disabled.
                click.secho('"{}" references an existing command. I am '
                            'ignoring the alias definition.'.format(cmd_name),
                            fg='yellow')
            else:
                # If the command references a configured alias, retrieve it
                # from the configuration.
                alias = self.get_config().get('alias', cmd_name)
                args = args[1:]
                if alias.startswith('!'):
                    cmd = SubprocessExecuter('')
                    additional_args = ' '.join(shlex.quote(a) for a in args)
                    return '', cmd, [alias[1:] + ' ' + additional_args]
                else:
                    args = shlex.split(alias) + args

        return super().resolve_command(ctx, args)

    def get_command(self, ctx, name):
        if name in self.get_configured_commands():
            path = self.get_config().get('commands', name)
            module_path, attr_name = path.rsplit('.', 1)
            module = importlib.import_module(module_path)
            return getattr(module, attr_name)
        else:
            return super().get_command(ctx, name)


@click.command(context_settings=CONTEXT_SETTINGS, cls=ConfigurableLoader)
@click.version_option(version=__version__, message='%(prog)s %(version)s')
@click.option('-d', '--debug/--no-debug', default=False,
              help=('Drop into the debugger if the command execution raises '
                    'an exception.'))
@click.option('--help-all', is_flag=True, is_eager=True, expose_value=False,
              callback=ConfigurableLoader.show_help_all,
              help='Show this message including hidden subcommands.')
@click.pass_context
def main(ctx, debug):
    # TODO: Enable this using a command line switch
    # import logging
    # logging.basicConfig(level=logging.DEBUG)

    if debug:
        def exception_handler(type, value, traceback):
            click.secho('\nAn exception occurred while executing the '
                        'requested command:', fg='red')
            hr(fg='red')
            sys.__excepthook__(type, value, traceback)

            click.secho('\nAs requested I will now drop you inside an '
                        'interactive debugging session:', fg='red')
            hr(fg='red')

            pdb.post_mortem(traceback)
        sys.excepthook = exception_handler

    try:
        integration_helper = ShellIntegrationHelper(
            os.environ['LANCET_SHELL_HELPER'])
    except KeyError:
        integration_helper = WarnIntegrationHelper()

    if os.path.exists(PROJECT_CONFIG):
        config = load_config(PROJECT_CONFIG)
    else:
        config = load_config()

    ctx.obj = Lancet(config, integration_helper)
    ctx.obj.call_on_close = ctx.call_on_close
    ctx.call_on_close(integration_helper.close)


@main.command()
def _setup_helper():
    """Print the shell integration code."""
    base = os.path.abspath(os.path.dirname(__file__))
    helper = os.path.join(base, 'helper.sh')
    with open(helper) as fh:
        click.echo(fh.read())


@main.command()
@click.pass_context
def _commands(ctx):
    """Prints a list of commands for shell completion hooks."""
    ctx = ctx.parent
    ctx.show_hidden_subcommands = False
    main = ctx.command

    for subcommand in main.list_commands(ctx):
        cmd = main.get_command(ctx, subcommand)
        if cmd is None:
            continue
        help = cmd.short_help or ''
        click.echo('{}:{}'.format(subcommand, help))


@main.command()
@click.argument('command_name', metavar='command', required=False)
@click.pass_context
def _arguments(ctx, command_name=None):
    """Prints a list of arguments for shell completion hooks.

    If a command name is given, returns the arguments for that subcommand.
    The command name has to refer to a command; aliases are not supported.
    """
    ctx = ctx.parent
    main = ctx.command
    if command_name:
        command = main.get_command(ctx, command_name)
        if not command:
            return
    else:
        command = main

    types = ['option', 'argument']
    all_params = sorted(command.get_params(ctx),
                        key=lambda p: types.index(p.param_type_name))

    def get_name(param):
        return max(param.opts, key=len)

    for param in all_params:
        if param.param_type_name == 'option':
            option = get_name(param)
            same_dest = [get_name(p) for p in all_params
                         if p.name == param.name]
            if same_dest:
                option = '({})'.format(' '.join(same_dest)) + option
            if param.help:
                option += '[{}]'.format(param.help or '')
            if not param.is_flag:
                option += '=:( )'
            click.echo(option)
        elif param.param_type_name == 'argument':
            option = get_name(param)
            click.echo(':{}'.format(option))


@main.command()
@click.argument('shell', required=False)
@click.pass_context
def _autocomplete(ctx, shell):
    """Print the shell autocompletion code."""
    if not shell:
        shell = os.environ.get('SHELL', '')
        shell = os.path.basename(shell).lower()
    if not shell:
        click.secho('Your shell could not be detected, please pass its name '
                    'as the argument.', fg='red')
        ctx.exit(-1)

    base = os.path.abspath(os.path.dirname(__file__))
    autocomplete = os.path.join(base, 'autocomplete', '{}.sh'.format(shell))

    if not os.path.exists(autocomplete):
        click.secho('Autocompletion for your shell ({}) is currently not '
                    'supported.', fg='red')
        ctx.exit(-1)

    with open(autocomplete) as fh:
        click.echo(fh.read())


# TODO:
# * review
#     pull
#     ci-status
#     pep8
#     diff
#     mergeability (rebase is of the submitter responsibility)
# * merge
#     pull, merge, delete
# * issues
#     list all open/assigned issues (or by filter)
# * comment
#     adds a comment to the currently active issue



import re

import click

from lancet.utils import hr


@click.command()
@click.argument('query', required=False)
@click.pass_obj
def projects(lancet, query):
    """List Harvest projects, optionally filtered with a regexp."""
    projects = lancet.timer.projects()

    if query:
        regexp = re.compile(query, flags=re.IGNORECASE)

        def match(project):
            match = regexp.search(project['name'])
            if match is None:
                return False
            project['match'] = match
            return True
        projects = (p for p in projects if match(p))

    for project in sorted(projects, key=lambda p: p['name'].lower()):
        name = project['name']

        if 'match' in project:
            m = project['match']
            s, e = m.start(), m.end()
            match = click.style(name[s:e], fg='green')
            name = name[:s] + match + name[e:]

        click.echo('{:>9d} {} {}'.format(
            project['id'], click.style('‣', fg='yellow'), name))


@click.command()
@click.argument('project_id', type=int)
@click.pass_obj
def tasks(lancet, project_id):
    """List Harvest tasks for the given project ID."""
    projects = lancet.timer.projects()

    for project in projects:
        if project['id'] == project_id:
            click.echo('{:>9d} {} {}'.format(
                project['id'], click.style('‣', fg='yellow'), project['name']))
            hr()

            for task in project['tasks']:
                click.echo('{:>9d} {} {}'.format(
                    task['id'], click.style('‣', fg='yellow'), task['name']))
            break



from shlex import quote

import click


@click.command()
@click.option('-p', '--print/--exec', 'print_cmd', default=False,
              help='Print the command instead of executing it.')
@click.argument('environment')
@click.pass_obj
def ssh(lancet, print_cmd, environment):
    """
    SSH into the given environment, based on the dploi configuration.
    """
    namespace = {}

    with open(lancet.config.get('dploi', 'deployment_spec')) as fh:
        code = compile(fh.read(), 'deployment.py', 'exec')
        exec(code, {}, namespace)

    config = namespace['settings'][environment]
    host = '{}@{}'.format(config['user'], config['hosts'][0])
    cmd = ['ssh', '-p', str(config.get('port', 22)), host]

    if print_cmd:
        click.echo(' '.join(quote(s) for s in cmd))
    else:
        lancet.defer_to_shell(*cmd)



import os
import configparser

import click
import keyring

from ..settings import USER_CONFIG, PROJECT_CONFIG
from ..utils import taskstatus


@click.command()
@click.option('-f', '--force/--no-force', default=False)
@click.pass_context
def setup(ctx, force):
    """Wizard to create the user-level configuration file."""
    if os.path.exists(USER_CONFIG) and not force:
        click.secho(
            'An existing configuration file was found at "{}".\n'
            .format(USER_CONFIG),
            fg='red', bold=True
        )
        click.secho(
            'Please remove it before in order to run the setup wizard or use\n'
            'the --force flag to overwrite it.'
        )
        ctx.exit(1)

    click.echo('Address of the issue tracker (your JIRA instance). \n'
               'Normally in the form https://<company>.atlassian.net.')
    tracker_url = click.prompt('URL')
    tracker_user = click.prompt('Username for {}'.format(tracker_url))
    click.echo()

    click.echo('Address of the time tracker (your Harvest instance). \n'
               'Normally in the form https://<company>.harvestapp.com.')
    timer_url = click.prompt('URL')
    timer_user = click.prompt('Username for {}'.format(timer_url))
    click.echo()

    config = configparser.ConfigParser()

    config.add_section('tracker')
    config.set('tracker', 'url', tracker_url)
    config.set('tracker', 'username', tracker_user)

    config.add_section('harvest')
    config.set('harvest', 'url', timer_url)
    config.set('harvest', 'username', timer_user)

    with open(USER_CONFIG, 'w') as fh:
        config.write(fh)

    click.secho('Configuration correctly written to "{}".'
                .format(USER_CONFIG), fg='green')

    # TODO: Add wizard to setup shell integration


@click.command()
@click.option('-f', '--force/--no-force', default=False,
              help='Init even if .lancet already exists.')
@click.pass_context
def init(ctx, force):
    """Wizard to create a project-level configuration file."""
    if os.path.exists(PROJECT_CONFIG) and not force:
        click.secho(
            'An existing configuration file was found at "{}".\n'
            .format(PROJECT_CONFIG),
            fg='red', bold=True
        )
        click.secho(
            'Please remove it before in order to run the setup wizard or use\n'
            'the --force flag to overwrite it.'
        )
        ctx.exit(1)

    project_key = click.prompt('Project key on the issue tracker')
    base_branch = click.prompt('Integration branch', default='master')

    virtualenvs = ('.venv', '.env', 'venv', 'env')
    for p in virtualenvs:
        if os.path.exists(os.path.join(p, 'bin', 'activate')):
            venv = p
            break
    else:
        venv = ''
    venv_path = click.prompt('Path to virtual environment', default=venv)

    project_id = click.prompt('Project ID on Harvest', type=int)
    task_id = click.prompt('Task id on Harvest', type=int)

    config = configparser.ConfigParser()

    config.add_section('lancet')
    config.set('lancet', 'virtualenv', venv_path)

    config.add_section('tracker')
    config.set('tracker', 'default_project', project_key)

    config.add_section('harvest')
    config.set('harvest', 'project_id', str(project_id))
    config.set('harvest', 'task_id', str(task_id))

    config.add_section('repository')
    config.set('repository', 'base_branch', base_branch)

    with open(PROJECT_CONFIG, 'w') as fh:
        config.write(fh)

    click.secho('\nConfiguration correctly written to "{}".'
                .format(PROJECT_CONFIG), fg='green')


@click.command()
@click.argument('service', required=False)
@click.pass_obj
def logout(lancet, service):
    """Forget saved passwords for the web services."""
    if service:
        services = [service]
    else:
        services = ['tracker', 'harvest']

    for service in services:
        url = lancet.config.get(service, 'url')
        key = 'lancet+{}'.format(url)
        username = lancet.config.get(service, 'username')
        with taskstatus('Logging out from {}', url) as ts:
            if keyring.get_password(key, username):
                keyring.delete_password(key, username)
                ts.ok('Logged out from {}', url)
            else:
                ts.ok('Already logged out from {}', url)


@click.command()
@click.pass_obj
def _services(lancet):
    """List all currently configured services."""
    def get_services(config):
        for s in config.sections():
            if config.has_option(s, 'url'):
                if config.has_option(s, 'username'):
                    yield s

    for s in get_services(lancet.config):
        click.echo('{}[Logout from {}]'.format(s, lancet.config.get(s, 'url')))



import click
import github3
from giturlparse import parse as giturlparse

from ..utils import taskstatus, edit_template
from ..helpers import get_issue, get_transition, set_issue_status, get_branch


@click.command()
@click.option('--base', '-b', 'base_branch',
              help='Branch to make pull request to.')
@click.option('-s', '--stop-timer/--no-stop-timer', default=False,
              help='Stops the Harvest timer after creating the pull request.')
@click.option('-o', '--open-pr/--no-open-pr', default=False,
              help='Opens the link with the pull request.')
@click.pass_context
def pull_request(ctx, base_branch, open_pr, stop_timer):
    """Create a new pull request for this issue."""
    lancet = ctx.obj

    username = lancet.config.get('tracker', 'username')
    review_status = lancet.config.get('tracker', 'review_status')
    remote_name = lancet.config.get('repository', 'remote_name')

    if not base_branch:
        base_branch = lancet.config.get('repository', 'base_branch')

    # Get the issue
    issue = get_issue(lancet)

    transition = get_transition(ctx, lancet, issue, review_status)

    # Get the working branch
    branch = get_branch(lancet, issue, create=False)

    with taskstatus('Checking pre-requisites') as ts:
        if not branch:
            ts.abort('No working branch found')

        assignee = issue.fields.assignee
        if not assignee or assignee.name != username:
            ts.abort('Issue currently not assigned to you')

        # TODO: Check mergeability

    # TODO: Check remote status (PR does not already exist)

    # Push to remote
    with taskstatus('Pushing to "{}"', remote_name) as ts:
        remote = lancet.repo.lookup_remote(remote_name)
        if not remote:
            ts.abort('Remote "{}" not found', remote_name)

        remote.credentials = lancet.repo.get_credentials_for_remote(remote)
        remote.push([branch.name])

        ts.ok('Pushed latest changes to "{}"', remote_name)

    # Create pull request
    with taskstatus('Creating pull request') as ts:
        p = giturlparse(remote.url)
        gh_repo = lancet.github.repository(p.owner, p.repo)

        template_path = lancet.config.get('repository', 'pr_template')
        message = edit_template(template_path, issue=issue)

        if not message:
            ts.abort('You didn\'t provide a title for the pull request')

        title, body = message.split('\n', 1)
        title = title.strip()

        if not title:
            ts.abort('You didn\'t provide a title for the pull request')

        try:
            pr = gh_repo.create_pull(title, base_branch, branch.branch_name,
                                     body.strip('\n'))
        except github3.GitHubError as e:
            if len(e.errors) == 1:
                error = e.errors[0]
                if 'pull request already exists' in error['message']:
                    ts.ok('Pull request does already exist')
                else:
                    ts.abort('Could not create pull request ({})',
                             error['message'])
            else:
                raise
        else:
            ts.ok('Pull request created at {}', pr.html_url)

    # Update issue
    set_issue_status(lancet, issue, review_status, transition)

    # TODO: Post to activity stream on JIRA
    # TODO: Post to HipChat?

    # Stop harvest timer
    if stop_timer:
        with taskstatus('Pausing harvest timer') as ts:
            lancet.timer.pause()
            ts.ok('Harvest timer paused')

    # Open the pull request page in the browser if requested
    if open_pr:
        click.launch(pr.html_url)


@click.command()
@click.option('-f', '--force/--no-stop-force', default=False,
              help='Creates a new branch if it does not exist yet.')
@click.argument('issue')
@click.pass_obj
def checkout(lancet, force, issue):
    """
    Checkout the branch for the given issue.

    It is an error if the branch does no exist yet.
    """
    issue = get_issue(lancet, issue)

    # Get the working branch
    branch = get_branch(lancet, issue, create=force)

    with taskstatus('Checking out working branch') as ts:
        if not branch:
            ts.abort('Working branch not found')
        lancet.repo.checkout(branch.name)
        ts.ok('Checked out "{}"', branch.name)



import click

from ..helpers import get_issue


@click.command()
@click.argument('issue', required=False)
@click.pass_obj
def browse(lancet, issue):
    """
    Open the issue tracker page for the given issue in your default browser.

    If no issue is provided, the one linked to the current branch is assumed.
    """
    click.launch(get_issue(lancet, issue).permalink())



from collections import OrderedDict
import subprocess

import click
import pygit2
from jinja2 import Template

from ..utils import content_from_path


@click.command()
@click.argument('output', type=click.File('wb'))
@click.pass_obj
def contributors(lancet, output):
    """
    List all contributors visible in the git history.
    """
    sorting = pygit2.GIT_SORT_TIME | pygit2.GIT_SORT_REVERSE
    commits = lancet.repo.walk(lancet.repo.head.target, sorting)
    contributors = ((c.author.name, c.author.email) for c in commits)
    contributors = OrderedDict(contributors)

    template_content = content_from_path(
        lancet.config.get('packaging', 'contributors_template'))
    template = Template(template_content)
    output.write(template.render(contributors=contributors).encode('utf-8'))


@click.command()
@click.argument('version', required=False)
@click.pass_obj
def tag_version(lancet, version):
    # NOTE: We're using a subprocess instead of interpreting the setup.py file
    # as the python version of the package may differ from the one which
    # executes lancet.
    if not version:
        version = subprocess.check_output(['python', 'setup.py', '--version'])
        version = version.strip().decode('ascii')

    name = subprocess.check_output(['python', 'setup.py', '--name'])
    name = name.strip().decode('ascii')

    tag_name = lancet.config.get('packaging', 'version_tag_name').format(
        name=name, version=version)
    tag_message = lancet.config.get('packaging', 'version_tag_message').format(
        name=name, version=version)

    click.echo('I am tagging the current commit with the version {}.'.format(
        click.style(tag_name, fg='green'),
    ))

    if click.confirm('Do you want to continue?'):
        lancet.repo.create_tag(
            tag_name,
            lancet.repo.head.target,
            pygit2.GIT_OBJ_COMMIT,
            lancet.repo.default_signature,
            tag_message
        )
